@@ -9,6 +9,7 @@ import (
 	"crypto/sha256"
 	"fmt"
 	"go/ast"
+	"go/build"
 	"go/parser"
 	"go/token"
 	"go/types"
@@ -23,12 +24,14 @@ import (
 	"golang.org/x/sync/errgroup"
 	"golang.org/x/tools/go/ast/astutil"
 	"golang.org/x/tools/gopls/internal/cache/metadata"
+	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/cache/typerefs"
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/filecache"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
+	"golang.org/x/tools/gopls/internal/util/slices"
 	"golang.org/x/tools/internal/analysisinternal"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/event/tag"
@@ -204,9 +207,17 @@ func (s *Snapshot) resolveImportGraph() (*importGraph, error) {
 
 	openPackages := make(map[PackageID]bool)
 	for _, fh := range s.Overlays() {
-		mps, err := s.MetadataForFile(ctx, fh.URI())
-		if err != nil {
-			return nil, err
+		// golang/go#66145: don't call MetadataForFile here. This function, which
+		// builds a shared import graph, is an optimization. We don't want it to
+		// have the side effect of triggering a load.
+		//
+		// In the past, a call to MetadataForFile here caused a bunch of
+		// unnecessary loads in multi-root workspaces (and as a result, spurious
+		// diagnostics).
+		g := s.MetadataGraph()
+		var mps []*metadata.Package
+		for _, id := range g.IDs[fh.URI()] {
+			mps = append(mps, g.Packages[id])
 		}
 		metadata.RemoveIntermediateTestVariants(&mps)
 		for _, mp := range mps {
@@ -675,7 +686,7 @@ func (b *typeCheckBatch) checkPackageForImport(ctx context.Context, ph *packageH
 	// Parse the compiled go files, bypassing the parse cache as packages checked
 	// for import are unlikely to get cache hits. Additionally, we can optimize
 	// parsing slightly by not passing parser.ParseComments.
-	pgfs := make([]*ParsedGoFile, len(ph.localInputs.compiledGoFiles))
+	pgfs := make([]*parsego.File, len(ph.localInputs.compiledGoFiles))
 	{
 		var group errgroup.Group
 		// Set an arbitrary concurrency limit; we want some parallelism but don't
@@ -1271,7 +1282,7 @@ func (s *Snapshot) typerefData(ctx context.Context, id PackageID, imports map[Im
 		bug.Reportf("internal error reading typerefs data: %v", err)
 	}
 
-	pgfs, err := s.view.parseCache.parseFiles(ctx, token.NewFileSet(), ParseFull&^parser.ParseComments, true, cgfs...)
+	pgfs, err := s.view.parseCache.parseFiles(ctx, token.NewFileSet(), parsego.Full&^parser.ParseComments, true, cgfs...)
 	if err != nil {
 		return nil, err
 	}
@@ -1453,9 +1464,10 @@ func (b *typeCheckBatch) checkPackage(ctx context.Context, ph *packageHandle) (*
 	defer done()
 
 	pkg := &syntaxPackage{
-		id:    inputs.id,
-		fset:  b.fset, // must match parse call below
-		types: types.NewPackage(string(inputs.pkgPath), string(inputs.name)),
+		id:         inputs.id,
+		fset:       b.fset, // must match parse call below
+		types:      types.NewPackage(string(inputs.pkgPath), string(inputs.name)),
+		typesSizes: inputs.sizes,
 		typesInfo: &types.Info{
 			Types:      make(map[ast.Expr]types.TypeAndValue),
 			Defs:       make(map[*ast.Ident]types.Object),
@@ -1471,11 +1483,11 @@ func (b *typeCheckBatch) checkPackage(ctx context.Context, ph *packageHandle) (*
 	// Collect parsed files from the type check pass, capturing parse errors from
 	// compiled files.
 	var err error
-	pkg.goFiles, err = b.parseCache.parseFiles(ctx, b.fset, ParseFull, false, inputs.goFiles...)
+	pkg.goFiles, err = b.parseCache.parseFiles(ctx, b.fset, parsego.Full, false, inputs.goFiles...)
 	if err != nil {
 		return nil, err
 	}
-	pkg.compiledGoFiles, err = b.parseCache.parseFiles(ctx, b.fset, ParseFull, false, inputs.compiledGoFiles...)
+	pkg.compiledGoFiles, err = b.parseCache.parseFiles(ctx, b.fset, parsego.Full, false, inputs.compiledGoFiles...)
 	if err != nil {
 		return nil, err
 	}
@@ -1620,11 +1632,8 @@ func (b *typeCheckBatch) typesConfig(ctx context.Context, inputs typeCheckInputs
 
 	if inputs.goVersion != "" {
 		goVersion := "go" + inputs.goVersion
-		// types.NewChecker panics if GoVersion is invalid. An unparsable mod
-		// file should probably stop us before we get here, but double check
-		// just in case.
-		if goVersionRx.MatchString(goVersion) {
-			typesinternal.SetGoVersion(cfg, goVersion)
+		if validGoVersion(goVersion) {
+			cfg.GoVersion = goVersion
 		}
 	}
 
@@ -1634,6 +1643,26 @@ func (b *typeCheckBatch) typesConfig(ctx context.Context, inputs typeCheckInputs
 	return cfg
 }
 
+// validGoVersion reports whether goVersion is a valid Go version for go/types.
+// types.NewChecker panics if GoVersion is invalid.
+//
+// Note that, prior to go1.21, go/types required exactly two components to the
+// version number. For example, go types would panic with the Go version
+// go1.21.1. validGoVersion handles this case when built with go1.20 or earlier.
+func validGoVersion(goVersion string) bool {
+	if !goVersionRx.MatchString(goVersion) {
+		return false // malformed version string
+	}
+
+	// TODO(rfindley): remove once we no longer support building gopls with Go
+	// 1.20 or earlier.
+	if !slices.Contains(build.Default.ReleaseTags, "go1.21") && strings.Count(goVersion, ".") >= 2 {
+		return false // unsupported patch version
+	}
+
+	return true
+}
+
 // depsErrors creates diagnostics for each metadata error (e.g. import cycle).
 // These may be attached to import declarations in the transitive source files
 // of pkg, or to 'requires' declarations in the package's go.mod file.
@@ -1670,12 +1699,12 @@ func depsErrors(ctx context.Context, snapshot *Snapshot, mp *metadata.Package) (
 
 	// Build an index of all imports in the package.
 	type fileImport struct {
-		cgf *ParsedGoFile
+		cgf *parsego.File
 		imp *ast.ImportSpec
 	}
 	allImports := map[string][]fileImport{}
 	for _, uri := range mp.CompiledGoFiles {
-		pgf, err := parseGoURI(ctx, snapshot, uri, ParseHeader)
+		pgf, err := parseGoURI(ctx, snapshot, uri, parsego.Header)
 		if err != nil {
 			return nil, err
 		}
@@ -1822,6 +1851,12 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, errs []types.Error, linkTarget
 				// report.
 				continue
 			}
+
+			// Invariant: both start and end are IsValid.
+			if !end.IsValid() {
+				panic("end is invalid")
+			}
+
 			posn := safetoken.StartPosition(e.Fset, start)
 			if !posn.IsValid() {
 				// All valid positions produced by the type checker should described by
@@ -1847,10 +1882,34 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, errs []types.Error, linkTarget
 				}
 				continue
 			}
-			if !end.IsValid() || end == start {
+
+			// debugging #65960
+			//
+			// At this point, we know 'start' IsValid, and
+			// StartPosition(start) worked (with e.Fset).
+			//
+			// If the asserted condition is true, 'start'
+			// is also in range for pgf.Tok, which means
+			// the PosRange failure must be caused by 'end'.
+			if pgf.Tok != e.Fset.File(start) {
+				bug.Reportf("internal error: inconsistent token.Files for pos")
+			}
+
+			if end == start {
 				// Expand the end position to a more meaningful span.
 				end = analysisinternal.TypeErrorEndPos(e.Fset, pgf.Src, start)
+
+				// debugging #65960
+				if _, err := safetoken.Offset(pgf.Tok, end); err != nil {
+					bug.Reportf("TypeErrorEndPos returned invalid end: %v", err)
+				}
+			} else {
+				// debugging #65960
+				if _, err := safetoken.Offset(pgf.Tok, end); err != nil {
+					bug.Reportf("ReadGo116ErrorData returned invalid end: %v", err)
+				}
 			}
+
 			rng, err := pgf.Mapper.PosRange(pgf.Tok, start, end)
 			if err != nil {
 				bug.Reportf("internal error: could not compute pos to range for %v: %v", e, err)
