@@ -10,6 +10,7 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
+	"go/ast"
 	"io"
 	"os"
 	"path/filepath"
@@ -21,6 +22,7 @@ import (
 	"sync"
 
 	"golang.org/x/mod/modfile"
+	"golang.org/x/telemetry/counter"
 	"golang.org/x/tools/go/ast/astutil"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/metadata"
@@ -32,7 +34,6 @@ import (
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/protocol/command"
 	"golang.org/x/tools/gopls/internal/settings"
-	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/vulncheck"
 	"golang.org/x/tools/gopls/internal/vulncheck/scan"
@@ -80,7 +81,13 @@ func (*commandHandler) AddTelemetryCounters(_ context.Context, args command.AddT
 		return fmt.Errorf("Names and Values must have the same length")
 	}
 	// invalid counter update requests will be silently dropped. (no audience)
-	telemetry.AddForwardedCounters(args.Names, args.Values)
+	for i, n := range args.Names {
+		v := args.Values[i]
+		if n == "" || v < 0 {
+			continue
+		}
+		counter.Add("fwd/"+n, v)
+	}
 	return nil
 }
 
@@ -496,6 +503,88 @@ func (c *commandHandler) Test(ctx context.Context, uri protocol.DocumentURI, tes
 	})
 }
 
+func (c *commandHandler) Doc(ctx context.Context, loc protocol.Location) error {
+	return c.run(ctx, commandConfig{
+		progress: "", // the operation should be fast
+		forURI:   loc.URI,
+	}, func(ctx context.Context, deps commandDeps) error {
+		pkg, pgf, err := golang.NarrowestPackageForFile(ctx, deps.snapshot, loc.URI)
+		if err != nil {
+			return err
+		}
+
+		// When invoked from a _test.go file, show the
+		// documentation of the package under test.
+		pkgpath := pkg.Metadata().PkgPath
+		if pkg.Metadata().ForTest != "" {
+			pkgpath = pkg.Metadata().ForTest
+		}
+
+		// Start web server.
+		web, err := c.s.getWeb()
+		if err != nil {
+			return err
+		}
+
+		// Compute fragment (e.g. "#Buffer.Len") based on
+		// enclosing top-level declaration, if exported.
+		var fragment string
+		pos, err := pgf.PositionPos(loc.Range.Start)
+		if err != nil {
+			return err
+		}
+		path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos)
+		if n := len(path); n > 1 {
+			switch decl := path[n-2].(type) {
+			case *ast.FuncDecl:
+				if decl.Name.IsExported() {
+					// e.g. "#Println"
+					fragment = decl.Name.Name
+
+					// method?
+					if decl.Recv != nil && len(decl.Recv.List) > 0 {
+						recv := decl.Recv.List[0].Type
+						if star, ok := recv.(*ast.StarExpr); ok {
+							recv = star.X // *N -> N
+						}
+						if id, ok := recv.(*ast.Ident); ok && id.IsExported() {
+							// e.g. "#Buffer.Len"
+							fragment = id.Name + "." + fragment
+						} else {
+							fragment = ""
+						}
+					}
+				}
+
+			case *ast.GenDecl:
+				// path=[... Spec? GenDecl File]
+				for _, spec := range decl.Specs {
+					if n > 2 && spec == path[n-3] {
+						var name *ast.Ident
+						switch spec := spec.(type) {
+						case *ast.ValueSpec:
+							// var, const: use first name
+							name = spec.Names[0]
+						case *ast.TypeSpec:
+							name = spec.Name
+						}
+						if name != nil && name.IsExported() {
+							fragment = name.Name
+						}
+						break
+					}
+				}
+			}
+		}
+
+		// Direct the client to open the /pkg page.
+		url := web.pkgURL(deps.snapshot.View(), pkgpath, fragment)
+		openClientBrowser(ctx, c.s.client, url)
+
+		return nil
+	})
+}
+
 func (c *commandHandler) RunTests(ctx context.Context, args command.RunTestsArgs) error {
 	return c.run(ctx, commandConfig{
 		async:       true,
@@ -745,6 +834,7 @@ func (s *server) getUpgrades(ctx context.Context, snapshot *cache.Snapshot, uri
 	stdout, err := snapshot.RunGoCommandDirect(ctx, cache.Normal|cache.AllowNetwork, &gocommand.Invocation{
 		Verb:       "list",
 		Args:       append([]string{"-m", "-u", "-json"}, modules...),
+		ModFlag:    "readonly", // necessary when vendor is present (golang/go#66055)
 		WorkingDir: filepath.Dir(uri.Path()),
 	})
 	if err != nil {
@@ -814,7 +904,7 @@ func (c *commandHandler) ListImports(ctx context.Context, args command.URIArg) (
 		if err != nil {
 			return err
 		}
-		pgf, err := deps.snapshot.ParseGo(ctx, fh, parsego.ParseHeader)
+		pgf, err := deps.snapshot.ParseGo(ctx, fh, parsego.Header)
 		if err != nil {
 			return err
 		}
@@ -1218,6 +1308,9 @@ func openClientBrowser(ctx context.Context, cli protocol.Client, url protocol.UR
 
 // openClientEditor causes the LSP client to open the specified document
 // and select the indicated range.
+//
+// Note that VS Code 1.87.2 doesn't currently raise the window; this is
+// https://github.com/microsoft/vscode/issues/207634
 func openClientEditor(ctx context.Context, cli protocol.Client, loc protocol.Location) {
 	showDocumentImpl(ctx, cli, protocol.URI(loc.URI), &loc.Range)
 }
